# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Simple unit test for configOverrides."""

import copy
import pickle
import unittest

from lsst.daf.butler import DatasetRef, DatasetType, DimensionUniverse
from lsst.pipe.base import DeferredDatasetRef
from lsst.utils.tests import MemoryTestCase, init


class DeferredDatasetRefTestCase(unittest.TestCase):
    """A test case for  DeferredDatasetRef"""

    def test_checkCopy(self):
        dsr = DatasetRef(
            datasetType=DatasetType(
                "pvi", dimensions="visit", storageClass="dict", universe=DimensionUniverse()
            ),
            dataId=dict(visit=10, instrument="test"),
            run="na",
        )
        ddsr = DeferredDatasetRef(dsr)
        ddsr_copy = copy.deepcopy(ddsr)
        assert isinstance(ddsr_copy, DeferredDatasetRef)

        round_trip = pickle.loads(pickle.dumps(ddsr))
        assert isinstance(round_trip, DeferredDatasetRef)


class MyMemoryTestCase(MemoryTestCase):
    """Check for memory leaks."""


def setup_module(module):
    """Configure pytest."""
    init()


if __name__ == "__main__":
    init()
    unittest.main()
